from __future__ import annotations

import json
import re
from hashlib import md5
from logging import WARNING, DEBUG, INFO, Logger
from pathlib import Path
from typing import Optional, Generic, Any, Dict, cast, TypeVar, Iterable

from persistable.data import PersistableParams
from persistable.exceptions import ExplainedNotImplementedError
from persistable.io import FileIO, PickleFileIO, PayloadTypeT
from persistable.logging import get_logger


def _camel_to_snake(name: str) -> str:
    name = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", name).lower()


PersistableParamsT = TypeVar("PersistableParamsT", bound=PersistableParams)


class Persistable(Generic[PayloadTypeT, PersistableParamsT]):
    """
    A payload wrapper for simple parameter based generation, persisting, and loading.

    The Persistable class is the centerpiece of a lightweight framework for quickly developing data and model pipelines.
    Pipelines are clearly parameterized, and parameter based persisting and loading means you'll never lose track of
    your results.  If you already ran a long running persistable pipline (i.e. generated a persistable payload), it can
    be reloaded with ease so it doesn't have to run again unless you explicitly want it to.
    """

    def __init__(
        self,
        data_dir: Path,
        params: PersistableParamsT,
        tracked_persistable_dependencies: Optional[Iterable[Persistable[Any, Any]]],
        *,
        payload_name: Optional[str] = None,
        payload_io: Optional[FileIO[PayloadTypeT]] = None,
        payload_file_suffix: str = ".persistable",
        verbose: bool = False,
        logger: Optional[Logger] = None,
    ) -> None:
        """
        Contstructor

        Args:
            data_dir (Path): Directory where persistable payloads will be persisted to and/or loaded from.
            params (PersistableParamsT): Parameters defining the pipeline and parameter based persisting and loading.
                These should be a dataclass like object that inherits from PersistableParams.
            tracked_persistable_dependencies (Optional[Iterable[Persistable[Any, Any]]]): Persistable dependencies with
                parameters that should be tracked for the purpose of persisting and loading.  If the payload depends on
                access to other persistable payloads (regardless of if they've been generated or loaded), those
                persistable objects should be provided here.  Otherwise pass None.
            payload_name (Optional[str]): Name the payload.  Useful for recognizing persisted data-files and log-files.
                For the purpose of parameter base persisting and loading, the payload_name is also used.  When no
                payload_name is provided, the Persistable class-name in camel-case is used.
            payload_io (Optional[FileIO[PayloadTypeT]]): IO with input
            payload_file_suffix ():
            verbose ():
            logger ():
        """
        if not data_dir.is_dir():
            data_dir.mkdir(parents=True)
        self.data_dir = data_dir
        self.params = params
        if tracked_persistable_dependencies is None:
            tracked_persistable_dependencies = []
        self.tracked_persistable_dependencies = tracked_persistable_dependencies
        if payload_name is None:
            payload_name = _camel_to_snake(self.__class__.__name__)
        self.payload_name = payload_name
        if payload_io is None:
            payload_io = PickleFileIO()
        self.payload_io = payload_io
        self.payload_file_suffix = payload_file_suffix

        console_level: int
        if verbose:
            console_level = INFO
        else:
            console_level = WARNING

        if logger is None:
            logger = get_logger(
                payload_name,
                file_loc=data_dir / f"{payload_name}.log",
                file_level=DEBUG,
                console_level=console_level,
            )

        self.logger = logger
        self.logger.info(f"---- NEW PERSISTABLE SESSION ---- ({data_dir})")
        self.logger.info(f"Payload named {payload_name}; Parameters set to {params}")

        self._payload: Optional[PayloadTypeT] = None

    @property
    def params_tree(self) -> Dict[str, Any]:
        return self.params.to_dict() | {
            persistable_obj.payload_name: persistable_obj.params.to_dict()
            for persistable_obj in self.tracked_persistable_dependencies
        }

    def generate(self, persist: bool = True, **untracked_payload_params: Any) -> None:
        """
        Generates payload and (by default) persist it.

        Parameters
        ----------
        persist                     : bool
            Default True, the payload is persisted
        untracked_payload_params    : dict
            These are helper parameters for generating an object that are not tracked.
            Generally these are not used.

        Returns
        -------

        """
        self.logger.info(f"Now generating {self.payload_name} payload...")
        self._payload = self._generate_payload(**untracked_payload_params)
        if persist:
            self.persist()

    def persist(self) -> None:

        """Persist both payload and parameters."""

        if self._payload is None:
            raise ValueError("Payload has not been generated.")

        self.payload_io.save(
            payload=self._payload,
            filepath=self.persist_filepath.with_suffix(self.payload_file_suffix),
        )
        with self.persist_filepath.with_suffix(".params.json").open("w") as params_file_handler:
            json.dump(self.params_tree, params_file_handler)

    def load(self, **untracked_payload_params: Any) -> None:
        """
        Loads persisted payload

        Parameters
        ----------
        untracked_payload_params    : dict
            Parameters not tracked by persistable that are only used to run the _post_load.
            Such scripts are useful if part of the payload cannot be persisted and needs to be recalculated
            at load time.

        Returns
        -------

        """
        self.logger.info(f"Now loading {self.payload_name} payload...")
        # ToDo - add find similar file functionality
        self._payload = self.payload_io.load(self.persist_filepath.with_suffix(self.payload_file_suffix))
        self._post_load()

    def load_generate(self, **untracked_payload_params: Any) -> None:
        """
        Like load() but executes the generate() method if load() fails due to a FileNotFoundError.

        Returns
        -------

        """
        try:
            self.load(**untracked_payload_params)
        except FileNotFoundError:
            self.logger.info("Loading payload failed, continuing to generate payload...")
            self.generate(**untracked_payload_params)

    @property
    def payload(self) -> PayloadTypeT:
        if self._payload is None:
            self.load_generate()
        return cast(PayloadTypeT, self._payload)

    def reset_payload(self) -> None:
        self._payload = None

    def _generate_payload(self, **untracked_payload_params: Any) -> PayloadTypeT:
        """
        Define here the algorithm for generating the payload
        based on self.params

        Parameters
        ----------
        untracked_payload_params    : dict
            Payload parameters that the user doesn't want to track (not persisted to file)

        Returns
        -------
        """

        raise ExplainedNotImplementedError(method_name=self._generate_payload.__name__)

    def _post_load(self) -> None:
        """
        Define here extra algorithmic steps to run after loading the payload.

        This is sometimes useful to add back components of the payload that are inefficient to persist or
        should not be persisted.

        Parameters
        ----------
        untracked_payload_params    : dict
            Payload parameters that the user doesn't want to track (not persisted to file)

        Returns
        -------

        """

    @property
    def persist_filepath(self) -> Path:
        params_tree_hex = md5(str({self.payload_name: self.params_tree}).encode()).hexdigest()
        return self.data_dir / f"{self.payload_name}({params_tree_hex})"



import logging
import sys
from pathlib import Path
from typing import Optional


def get_logger(
    name: str,
    file_loc: Optional[Path],
    console_level: Optional[int] = logging.DEBUG,
    file_level: Optional[int] = logging.DEBUG,
    format_str: str = "%(asctime)s - %(name)s - %(funcName)s - %(levelname)s - %(message)s",
    adapt_from_logger: Optional[logging.Logger] = None,
) -> logging.Logger:
    """Get logger utiltity.

    ToDo - clean up using a builder pattern.

    has defaults for everything. just enter
    >>> logger = get_logger(
    >>>     name="Package", file_loc=Path("path/to/log_files/"), console_level=logging.INFO, file_level=logging.DEBUG
    >>> )
    if you want to deactivate the console_level logger (on by default) enter:
    >>> logger = get_logger(
    >>>     name="Package", file_loc=Path("path/to/log_files/"), console_level=None, file_level=logging.DEBUG
    >>> )
    """
    # default values:
    formatter = logging.Formatter(format_str)

    logger: logging.Logger
    if adapt_from_logger:
        logger = adapt_from_logger
    else:
        logger = logging.getLogger(name)
        logger.setLevel(logging.DEBUG)  # set it to lowest possible, this needs to be done for logging to happen at all

    if console_level is not None:
        no_handlers = True
        for h in logger.handlers:
            if isinstance(h, logging.StreamHandler):
                h.setLevel(console_level)
                h.setFormatter(formatter)
                no_handlers = False
        if no_handlers:
            ch = logging.StreamHandler(sys.stdout)
            ch.setLevel(console_level)
            ch.setFormatter(formatter)
            logger.addHandler(ch)

    if file_loc is not None:
        if file_level is None:
            file_level = logging.DEBUG

        no_handlers = True
        for h in logger.handlers:
            if isinstance(h, logging.FileHandler):
                h.setLevel(file_level)
                h.setFormatter(formatter)
                no_handlers = False
        if no_handlers:
            fh = logging.FileHandler(file_loc, encoding="utf-8")
            fh.setLevel(file_level)
            fh.setFormatter(formatter)
            logger.addHandler(fh)

    return logger



from dataclasses import dataclass
from hashlib import md5
from typing import TypeVar, Generic

from persistable.io import DictEncodable

ParamTypeT = TypeVar("ParamTypeT")


class PersistableParam(Generic[ParamTypeT]):

    """Parameter wrapper for parameters with special meta options such as being ignored from file hash."""

    def __init__(self, param: ParamTypeT, exclude_from_file_hash: bool = False) -> None:
        self.param = param
        self.exclude_from_file_hash = exclude_from_file_hash


@dataclass
class PersistableParams(DictEncodable):
    def get_str_repr(self) -> str:
        return str(self.to_dict())

    def get_md5_hash(self) -> str:
        return md5(str(self).encode()).hexdigest()



import pickle
from abc import ABC, abstractmethod
from enum import Enum
from logging import getLogger
from pathlib import Path
from typing import TypeVar, Generic, Any, cast, Dict, Type, Set, Callable

import numpy as np

from persistable.exceptions import ExplainedNotImplementedError

PayloadTypeT = TypeVar("PayloadTypeT")


logger = getLogger(__name__)


class FileIO(Generic[PayloadTypeT], ABC):

    """FileIO Abstraction with load / save interface."""

    def load(self, filepath: Path, **kwargs: Any) -> PayloadTypeT:
        logger.debug("load `%s`", filepath)
        return self._load(filepath=filepath, **kwargs)

    def save(self, payload: PayloadTypeT, filepath: Path, create_folder: bool = True, **kwargs: Any) -> None:
        logger.debug("save %s to `%s`", payload.__class__.__name__, filepath)
        try:
            if create_folder and not filepath.parent.is_dir():
                filepath.parent.mkdir(parents=True)
            self._save(payload=payload, filepath=filepath, **kwargs)
        except Exception as exception:
            raise IOError(f"{payload.__class__.__name__} unable to save to `{filepath}`") from exception

    @abstractmethod
    def _load(self, filepath: Path, **kwargs: Any) -> PayloadTypeT:
        """Specific load from file implementation."""

    def _save(self, payload: PayloadTypeT, filepath: Path, **kwargs: Any) -> None:
        """Specific save to file implementation."""
        raise ExplainedNotImplementedError(self._save.__name__)


class PickleFileIO(FileIO[PayloadTypeT], Generic[PayloadTypeT]):
    def _load(self, filepath: Path, **kwargs: Any) -> PayloadTypeT:
        with filepath.open("rb") as file_handler:
            return cast(PayloadTypeT, pickle.load(file=file_handler, **kwargs))

    def _save(self, payload: PayloadTypeT, filepath: Path, **kwargs: Any) -> None:
        with filepath.open("wb") as file_handler:
            pickle.dump(obj=payload, file=file_handler, **kwargs)


class DictEncodable:

    """Mixin for encoding a python object to a dict.

    By default, this mixin will stringify Path and Enum objects.  To define more type based encoding methods, or
    override default behaviour, override the `_custom_type_encoder_map` property.
    """

    def to_dict(self) -> Dict[str, Any]:

        """Get a dictionary representation of self."""

        encoded_dict: Dict[str, Any] = dict()
        exluded_attributes = self._excluded_attributes
        custom_type_encoder_map = self._custom_type_encoder_map
        for attr_name, attr in self.__dict__.items():
            if attr_name in exluded_attributes:
                continue
            elif isinstance(attr, tuple(custom_type_encoder_map)):
                subclass_found_in_extra_dict_encodings: Type[Any] = [
                    class_ for class_ in attr.__class__.mro() if class_ in custom_type_encoder_map
                ][0]
                encoded_dict[attr_name] = custom_type_encoder_map[subclass_found_in_extra_dict_encodings](attr)
            elif isinstance(attr, Path):
                encoded_dict[attr_name] = str(attr)
            elif isinstance(attr, Enum):
                encoded_dict[attr_name] = attr.name
            elif isinstance(attr, DictEncodable):
                encoded_dict[attr_name] = attr.to_dict()
            else:
                encoded_dict[attr_name] = attr
        return encoded_dict

    @property
    def _excluded_attributes(self) -> Set[str]:
        """
        Define a set of self.__dict__ attributes that should be excluded from the dict encoder.

        For example, the self.__dict__ of a dataclass has an internal attribute `__initialised__`.  This is excluded
        by default.
        """
        return {
            "__initialised__",
        }

    @property
    def _custom_type_encoder_map(self) -> Dict[Type[Any], Callable[[Any], Any]]:
        """
        Define how attributes with certain types get mapped to dict.

        Note, this map also overrides default behaviour.

        Example:
            Override how Path objects are encoded (in this case, change "/" to "!_!"):
            >>> from pathlib import Path
            >>> return {
            >>>     Path: lambda path: str(path).replace("/", "!_!")
            >>> }
        """
        return dict()

    @staticmethod
    def dict_almost_equal(
        dict1: Dict[Any, Any],
        dict2: Dict[Any, Any],
        rtol: float = 1e-6,
        atol: float = 1e-6,
        verbose: bool = False,
    ) -> bool:
        if dict1.keys() != dict2.keys():
            return False

        msg = "{}: {}!={} (rtol={}, atol={})"
        for key, value in dict1.items():
            if isinstance(value, dict):
                if DictEncodable.dict_almost_equal(dict1[key], dict2[key], rtol=rtol, atol=atol):
                    continue
                else:
                    print(msg.format(key, dict1[key], dict2[key], rtol, atol)) if verbose else None
                    return False

            if isinstance(value, float):
                if cast(bool, np.isclose(dict1[key], dict2[key], rtol=rtol, atol=atol, equal_nan=True)):
                    continue
                else:
                    print(msg.format(key, dict1[key], dict2[key], rtol, atol)) if verbose else None
                    return False

            if isinstance(value, np.ndarray):
                if np.allclose(dict1[key], dict2[key], atol=atol, rtol=rtol):
                    continue
                else:
                    print(msg.format(key, dict1[key], dict2[key], rtol, atol)) if verbose else None
                    return False

            if dict1[key] == dict2[key]:
                continue

            print(msg.format(key, dict1[key], dict2[key], rtol, atol)) if verbose else None
            return False

        return True



from dataclasses import dataclass

import pytest

from persistable.data import PersistableParams
from persistable.io import DictEncodable


@dataclass
class DummyPersistableParams(PersistableParams):
    i: int
    f: float
    s: str


@dataclass
class PermutedDummyPersistableParams(PersistableParams):
    i: int
    s: str
    f: float


@dataclass
class DefaultsPersistableParams(PersistableParams):
    i: int = 10
    s: str = "hello"
    f: float = 12.24


class TestPersistableParams:
    def test_init(self) -> None:
        # GIVEN
        i = 10
        f = 12.2
        s = "test"

        # WHEN
        params = DummyPersistableParams(i=i, f=f, s=s)

        # THEN
        assert isinstance(params, PersistableParams)
        assert isinstance(params, DictEncodable)

    def test_to_dict(self) -> None:
        # GIVEN
        params = DefaultsPersistableParams()
        params_dict = params.to_dict()

        # THEN
        assert params_dict == dict(i=10, f=12.24, s="hello")

    @pytest.mark.skip(reason="Permuted params hash inequality feature not yet implements.")
    def test_hash_equality_when_params_permuted(self) -> None:
        # GIVEN
        i = 10
        f = 12.2
        s = "test"

        # WHEN
        params = DummyPersistableParams(i=i, f=f, s=s)
        permuted_params = PermutedDummyPersistableParams(i=i, f=f, s=s)

        # THEN
        assert params.get_md5_hash() == permuted_params.get_md5_hash()



import json
from dataclasses import dataclass
from logging import Logger
from pathlib import Path
from typing import Any, Dict, Optional

from persistable import Persistable
from persistable.data import PersistableParams
from persistable.io import PickleFileIO


@dataclass
class DummyPersistableParams(PersistableParams):
    a: int = 1
    b: str = "hello"


class DummyPersistable(Persistable[Dict[str, Any], DummyPersistableParams]):
    def _generate_payload(self, **untracked_payload_params: Any) -> Dict[str, Any]:
        return dict(a=1, b="test")


class DummyPersistableWithTrackedDependencies(Persistable[Dict[str, Any], DummyPersistableParams]):
    def __init__(
        self,
        data_dir: Path,
        params: DummyPersistableParams,
        *,
        dummy_persistable: DummyPersistable,
        verbose: bool = False,
        logger: Optional[Logger] = None,
    ) -> None:
        super().__init__(
            data_dir, params, tracked_persistable_dependencies=(dummy_persistable,), verbose=verbose, logger=logger
        )
        self.dummy_persistable = dummy_persistable

    def _generate_payload(self, **untracked_payload_params: Any) -> Dict[str, Any]:
        return dict(a=self.params.a, old=self.dummy_persistable.payload)


class TestPersistable:
    def test_init(self, tmp_path: Path) -> None:
        # GIVEN
        data_dir = tmp_path
        params = DummyPersistableParams()

        # WHEN
        persistable = Persistable(data_dir=data_dir, params=params, tracked_persistable_dependencies=None)

        # THEN
        assert isinstance(persistable, Persistable)
        assert persistable.data_dir == data_dir
        assert persistable.params == params
        assert persistable.payload_name == "persistable"
        assert isinstance(persistable.payload_io, PickleFileIO)

        assert persistable._payload is None

    def test_payload(self, tmp_path: Path) -> None:
        # GIVEN
        data_dir = tmp_path
        params = DummyPersistableParams()
        persistable = DummyPersistable(data_dir=data_dir, params=params, tracked_persistable_dependencies=None)

        # GIVEN expected artifact filepaths
        expected_persistable_filepath = persistable.persist_filepath.with_suffix(persistable.payload_file_suffix)
        expected_params_filepath = persistable.persist_filepath.with_suffix(".params.json")

        # WHEN
        payload = persistable.payload

        # THEN
        assert payload == dict(a=1, b="test")
        assert expected_persistable_filepath.exists()
        assert expected_params_filepath.exists()

        # WHEN
        with expected_params_filepath.open("r") as f:
            params_json = json.load(f)

        assert params_json == dict(a=1, b="hello")

    def test_tracked_persistable_dependencies(self, tmp_path: Path) -> None:
        # GIVEN
        data_dir = tmp_path
        params = DummyPersistableParams()
        dummy_persistable = DummyPersistable(data_dir=data_dir, params=params, tracked_persistable_dependencies=None)

        # WHEN persistable made from other persistables (here dummy persistable)
        from_other_persistables_persistable = DummyPersistableWithTrackedDependencies(
            data_dir=data_dir, params=params, dummy_persistable=dummy_persistable
        )

        # THEN the payload and params work as expected
        assert from_other_persistables_persistable.payload == dict(a=1, old=dummy_persistable.payload)
        assert from_other_persistables_persistable.params_tree == dict(
            a=1, b="hello", dummy_persistable=dict(a=1, b="hello")
        )

    def test_persist_filepath_determined_by_both_params_tree_and_payload_name(self, tmp_path: Path) -> None:
        # GIVEN
        data_dir = tmp_path
        params = DummyPersistableParams()
        dummy_persistable = DummyPersistable(data_dir=data_dir, params=params, tracked_persistable_dependencies=None)

        # WHEN and THEN
        assert dummy_persistable.persist_filepath == data_dir / "dummy_persistable(6a0f2e637a47f02428f19726be8541a1)"

        # WHEN payload_name changed
        dummy_persistable.payload_name = "another"

        # THEN persist filepath has changed (hash should be determined from params_tree and payload_name)
        assert dummy_persistable.persist_filepath == data_dir / "another(e15cab0e04c56c6deddb1ac7bc5e5956)"

        # WHEN params changed
        dummy_persistable.params.a += 10

        # THEN persist filepath has changed (hash should be determined from params_tree and payload_name)
        assert dummy_persistable.persist_filepath == data_dir / "another(bd9f250dac257114768e128ed4d9eb96)"



from distutils.core import setup
from typing import List

from setuptools import find_packages

install_requires: List[str] = ["numpy>=1.21.0"]
dev_requires: List[str] = [
    "pre-commit>=2.19.0",
    "pytest>=7.1.2",
]

setup(
    name="persistable",
    version="1.1.0",
    packages=find_packages(),
    url="https://github.com/aloosley/persistable",
    download_url="https://github.com/aloosley/persistable/archive/1.0.0.tar.gz",
    license="",
    author="Alex Loosley, Stephan Sahm",
    author_email="aloosley@alumni.brown.edu",
    description="An inheritable superclass with logging, and tools for persisting and "
    "loading models with parameter tracking",
    keywords=["persisting", "models", "pipeline"],
    install_requires=install_requires,
    extras_require=dict(dev=dev_requires),
    python_requires=">=3.9",
)
